<?php

namespace Drupal\varnish_file_purge\EventSubscriber;

use Drupal\Core\Entity\EntityEvent;
use Drupal\Core\Entity\EntityEvents;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\file\FileInterface;
use Drupal\varnish_file_purge\Services\VarnishFilePurger;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Class VarnishFilePurgeEventSubscriber.
 *
 * Creates a menu log on menu content entity events.
 */
class VarnishFilePurgeEventSubscriber implements EventSubscriberInterface {

  /**
   * Constructs a MenuEntityEventSubscriber object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   * @param \Drupal\varnish_file_purge\Services\VarnishFilePurger $varnishFilePurger
   *   The Varnish file purger.
   * @param \Drupal\Core\File\FileSystemInterface $fileSystem
   *   The file system.
   */
  public function __construct(
    protected EntityTypeManagerInterface $entityTypeManager,
    protected VarnishFilePurger $varnishFilePurger,
    protected FileSystemInterface $fileSystem,
  ) {}

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents(): array {
    $events[EntityEvents::INSERT] = 'onInsert';
    $events[EntityEvents::UPDATE] = 'onUpdate';
    $events[EntityEvents::DELETE] = 'onDelete';

    return $events;
  }

  /**
   * Creates a menu log when menu content is created.
   *
   * @param \Drupal\Core\Entity\EntityEvent $event
   *   The entity event.
   */
  public function onInsert(EntityEvent $event): void {
    $file = $event->getEntity();

    if (!$file instanceof FileInterface) {
      return;
    }

    if (!$this->varnishFilePurger->isConfigured()) {
      return;
    }

    $this->varnishFilePurger->purge($file);
  }

  /**
   * Creates a menu log when menu content is updated.
   *
   * @param \Drupal\Core\Entity\EntityEvent $event
   *   The entity event.
   */
  public function onUpdate(EntityEvent $event): void {
    $file = $event->getEntity();

    if (!$file instanceof FileInterface) {
      return;
    }

    if (!$this->varnishFilePurger->isConfigured()) {
      return;
    }

    /** @var \Drupal\file\FileInterface $original_file */
    $original_file = $file->original;

    if (!$original_file instanceof FileInterface) {
      return;
    }

    if ($this->isFileDifferent($file, $original_file)) {
      $this->varnishFilePurger->purge($file);
    }
  }

  /**
   * Track changes in the updated file by comparing it with the original.
   *
   * @param \Drupal\file\FileInterface $file
   *   The updated file.
   * @param \Drupal\file\FileInterface $original_file
   *   The original file.
   *
   * @return bool
   *   Whether the file is different from the original file.
   */
  protected function isFileDifferent(FileInterface $file, FileInterface $original_file): bool {
    // Compare file sizes and URIs as a quick check (optional).
    if ($original_file->getSize() !== $file->getSize()) {
      return TRUE;
    }

    // Compare file URIs to check if the file was moved or renamed.
    if ($file->getFileUri() != $original_file->getFileUri()) {
      return TRUE;
    }

    // The file URI is the same, but with the file_replace module,
    // content might still have changed.
    // Therefore, compare file contents using hashes.
    $original_file_path = $this->fileSystem->realpath($original_file->getFileUri());
    $updated_file_path = $this->fileSystem->realpath($file->getFileUri());

    // Generate hashes for both files.
    $original_file_hash = hash_file('sha256', $original_file_path);
    $updated_file_hash = hash_file('sha256', $updated_file_path);

    // Compare the hashes. If they differ, the file content has changed.
    if ($original_file_hash !== $updated_file_hash) {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * Creates a menu log when menu content is deleted.
   *
   * @param \Drupal\Core\Entity\EntityEvent $event
   *   The entity event.
   */
  public function onDelete(EntityEvent $event): void {
    $file = $event->getEntity();

    if (!$file instanceof FileInterface) {
      return;
    }

    if (!$this->varnishFilePurger->isConfigured()) {
      return;
    }

    $this->varnishFilePurger->purge($file);
  }

}
