<?php

namespace Drupal\varnish_file_purge\Services;

use Drupal\Core\Config\Config;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\file\FileInterface;
use Drupal\varnish_file_purge\Form\VarnishFilePurgeConfiguration;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Pool;
use Psr\Log\LoggerInterface;

/**
 * Helper class which sends the actual purge requests to Varnish.
 */
class VarnishFilePurger {

  /**
   * The configuration of the varnish file purger module.
   *
   * @var \Drupal\Core\Config\Config|\Drupal\Core\Config\ImmutableConfig
   */
  protected Config|ImmutableConfig $config;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface|\Drupal\Core\Logger\LoggerChannelInterface
   */
  protected LoggerInterface|LoggerChannelInterface $logger;

  /**
   * Constructs a VarnishFilePurgerHelper object.
   *
   * @param \GuzzleHttp\ClientInterface $client
   *   The HTTP client.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   The config factory.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $loggerChannelFactory
   *   The logger channel factory.
   */
  public function __construct(
    protected ClientInterface $client,
    protected ConfigFactoryInterface $configFactory,
    protected LoggerChannelFactoryInterface $loggerChannelFactory,
  ) {
    $this->config = $this->configFactory->get(VarnishFilePurgeConfiguration::CONFIG_KEY);
    $this->logger = $this->loggerChannelFactory->get('varnish_file_purge');
  }

  /**
   * Check if the varnish host is configured.
   *
   * @return bool
   *   Whether the host has been configured or not.
   */
  public function isConfigured(): bool {
    if (empty($this->config->get('host'))) {
      $this->logger->warning("Unable to purge: No Varnish host configured.");

      return FALSE;
    }

    return TRUE;
  }

  /**
   * Send purge requests to Varnish to purge all absolute URLs of a file.
   *
   * @param \Drupal\file\FileInterface $file
   *   The file whose URLs need to be purged.
   */
  public function purge(FileInterface $file): void {
    // Skip files that are in the temporary file stream.
    if (str_starts_with($file->getFileUri(), 'temporary://')) {
      return;
    }

    // Retrieve configuration settings.
    global $base_url;
    $domains = $this->config->get('domains') ?? $base_url;
    $domain_list = explode(',', $domains);
    $varnish_base_uri = $this->getBaseUri();
    $request_method = $this->config->get('request_method');
    $debug_mode = $this->config->get('debug');

    // Generate a generator function to yield asynchronous requests.
    $requests_generator = function () use ($file, $request_method, $domain_list, $varnish_base_uri, $debug_mode) {
      foreach ($domain_list as $domain) {
        // Options for the HTTP request, including the Host header.
        $options = [
          'base_uri' => $varnish_base_uri,
          'headers' => [
            'Host' => $domain,
          ],
        ];

        // Generate the absolute file URL to purge.
        $file_url = $file->createFileUrl();

        // Yield an asynchronous Guzzle request.
        yield function () use ($file_url, $request_method, $domain, $options, $debug_mode) {
          return $this->client->requestAsync($request_method, $file_url, $options)
            ->then(
            // Success callback: Log purging if debug mode is enabled.
              function () use ($file_url, $domain, $debug_mode) {
                if ($debug_mode) {
                  $this->logger->notice("URL purged: $file_url on domain: $domain");
                }
              },
              // Failure callback: Log the error.
              function ($reason) use ($file_url) {
                $message = $reason instanceof \Exception ? $reason->getMessage() : (string) $reason;
                $this->logger->error("URL not purged: $file_url. Error: {$message}");
              }
            );
        };
      }
    };

    // Use Guzzle's Pool to handle the asynchronous requests with concurrency.
    $pool = new Pool($this->client, $requests_generator(), [
      'concurrency' => 10,
    ]);

    // Wait for all requests to complete.
    $pool->promise()->wait();
  }

  /**
   * Returns the base URI of the varnish connection.
   *
   * @return string
   *   The base URI of the varnish connection.
   */
  protected function getBaseUri(): string {
    $base_uri = $this->config->get('scheme') ?? 'http://';
    $base_uri .= $this->config->get('host') ?? '';

    if (!empty($port = $this->config->get('port'))) {
      $base_uri .= ':' . $port;
    }

    return $base_uri;
  }

}
